
/* MODULE							HTAnnotate.c
**		HANDLE ANNOTATIONS ON SERVER SIDE
**
** AUTHORS:
**	AL	Ari Luotonen	luotonen@dxcern.cern.ch
**
** HISTORY:
**	17 Jan 94  AL	Written on a dark winter night from scratch.
**
** BUGS:
**
**
*/

#include <string.h>
#include <stdio.h>

#include "HTFile.h"
#include "HTUtils.h"
#include "HTParse.h"	/* HTUnEscape() */

#include "tcp.h"
#include "HTTCP.h"

#include "HTFormat.h"	/* HTInputSocket */
#include "HTAccess.h"
#include "HTRules.h"	/* HTBinDir, HTBINDIR */
#include "HTWriter.h"
#include "HTStream.h"
#include "HTRequest.h"
#include "HTDaemon.h"
#include "HTConfig.h"
#include "HTLog.h"


#ifdef USE_DIRENT		/* Set this for Sys V systems */
#define STRUCT_DIRENT struct dirent
#else
#define STRUCT_DIRENT struct direct
#endif


extern HTInputSocket * isoc;

PUBLIC char * HTGlobalAnnDir	= NULL;
PUBLIC char * HTDirwiseAnnDir	= NULL;
PUBLIC char * HTFilewiseAnnDir	= "_anndir";
PUBLIC char * HTAnnSuffix	= "_ann";


#ifndef VMS

PRIVATE char * get_ann_dir ARGS1(CONST char *, localname)
{
    char * ann_dir = NULL;
    char * p;

    if (!localname) return NULL;

    if (HTGlobalAnnDir || HTDirwiseAnnDir) {
	char * localpath = NULL;

	StrAllocCopy(localpath, localname);
	p = strrchr(localpath, '/');
	if (p) *p=0;
	else FREE(localpath);

	if (HTGlobalAnnDir) {
	    StrAllocCopy(ann_dir, HTGlobalAnnDir);
	    if (localpath) StrAllocCat(ann_dir, localpath);
	}
	else if (HTDirwiseAnnDir) {
	    if (localpath) StrAllocCopy(ann_dir, localpath);
	    StrAllocCat(ann_dir, "/");
	    StrAllocCat(ann_dir, HTDirwiseAnnDir);
	}
	FREE(localpath);
    }
    else if (HTFilewiseAnnDir) {
	StrAllocCopy(ann_dir, localname);
	p = strrchr(ann_dir, '.');
	if (p) p=0;
	StrAllocCat(ann_dir, HTFilewiseAnnDir);
    }
    return ann_dir;
}


PRIVATE char * get_ann_file ARGS1(HTRequest *, req)
{
    char * ann_dir;
    char * ann_base;
    char * ann_name;
    int ann_num = 0;
    int ann_len;
    char * p;

    DIR * dp;
    STRUCT_DIRENT * dirbuf;
    int baselen;
    int n;

    if (!req) return NULL;

    ann_dir = get_ann_dir(HTReqTranslated);
    p = strrchr(HTReqTranslated, '/');
    if (p) p++;
    else p = HTReqTranslated;
    StrAllocCopy(ann_base, p);
    StrAllocCat(ann_base, HTAnnSuffix);

    if (!ann_dir || !ann_base) return NULL;

    baselen = strlen(ann_base);
    dp = opendir(ann_dir);

    if (!dp) {
	CTRACE(stderr,"Error....... Can't open annotation directory %s\n",ann_dir);
	CTRACE(stderr,"Trying...... to create it: %s\n",ann_dir);
	if (mkdir(ann_dir, 0775) == -1) {
	    CTRACE(stderr, "Failed...... Couldn't create annotation dir %s\n",ann_dir);
	    FREE(ann_dir);
	    return NULL;
	}
	else {
	    CTRACE(stderr, "Ok.......... Annotation directory %s created\n", ann_dir);
	    dp = opendir(ann_dir);
	    if (!dp) {
		CTRACE(stderr, "ERROR....... BUT STILL DIRECTORY OPEN FAILED!!\n");
		return NULL;
	    }
	}
    }

    CTRACE(stderr, "Info........ Following annotations exists:\n");
    while ((dirbuf = readdir(dp))) {
	if (!dirbuf->d_ino) continue;	/* Not in use */
	if ((int)dirbuf->d_namlen > baselen  &&
	    !strncmp(dirbuf->d_name, ann_base, baselen)) {
	    p = dirbuf->d_name + baselen;
	    n = atoi(p);
	    if (n > ann_num) ann_num = n;
	    CTRACE(stderr, " #%d", n);
	}
    }
    closedir(dp);
    ann_num++;
    CTRACE(stderr,"\n............ End of annotations. New annotation is #%d\n",ann_num);

    ann_len = strlen(ann_dir) + strlen(ann_base) + 10;
    ann_name = (char*)malloc(ann_len);
    if (!ann_name) outofmem(__FILE__, "get_annotation_file");

    sprintf(ann_name, "%s/%s%d", ann_dir, ann_base, ann_num);

    return ann_name;
}
#endif /* not VMS */


PUBLIC int HTCopyBodyToFile ARGS2(HTRequest *,	req,
				  FILE *,	fp)
{
    int i;

    if (!req || !isoc || !fp) return -1;

    for(i=req->content_length; i>0; i--)
	fputc(HTInputSocket_getCharacter(isoc), fp);	/* @@ inefficient */

    return req->content_length - i;
}


PUBLIC int HTAnnotate ARGS1(HTRequest *, req)
{
#ifndef VMS 
    char * ann_file;
    FILE * fp;

    if (!req) return HT_LOADED;		/* What should I return?? */

    ann_file = get_ann_file(req);
    if (!ann_file) {
	CTRACE(stderr, "Error....... Unable to make annotation filename\n");
	return HT_LOADED;
    }

    fp = fopen(ann_file, "w");
    if (!fp) {
	CTRACE(stderr,"Error....... Can't open annot.file \"%s\" for writing\n",ann_file);
	return HT_LOADED;
    }

    CTRACE(stderr, "Info...... Writing annotation file \"%s\"\n", ann_file);
    HTCopyBodyToFile(req, fp);
    fclose(fp);

    HTLoadError(req, 200, "Ok, annotation created");

#endif /* not VMS */
    return HT_LOADED;
}

