
/* MODULE							HTPut.c
**		HANDLE PUT ON SERVER SIDE
**
** AUTHORS:
**	AL	Ari Luotonen	luotonen@dxcern.cern.ch
**
** HISTORY:
**	17 Jan 94  AL	First written.
**
** BUGS:
**
**
*/

#include <string.h>
#include <stdio.h>

#include "HTFile.h"
#include "HTUtils.h"
#include "HTParse.h"	/* HTUnEscape() */

#include "tcp.h"
#include "HTTCP.h"

#include "HTFormat.h"	/* HTInputSocket */
#include "HTAccess.h"
#include "HTRules.h"	/* HTBinDir, HTBINDIR */
#include "HTWriter.h"
#include "HTStream.h"

#include "HTRequest.h"
#include "HTDaemon.h"

#ifdef USE_DIRENT		/* Set this for Sys V systems */
#define STRUCT_DIRENT struct dirent
#else
#define STRUCT_DIRENT struct direct
#endif


extern HTInputSocket * isoc;

PUBLIC char * HTBackupSuffix = "_bak";


PRIVATE char * backup_name ARGS1(CONST char *, name)
{
    char * backup = NULL;

    if (!name) return NULL;

    StrAllocCopy(backup, name);
    StrAllocCat(backup, HTBackupSuffix);
    return backup;
}


PRIVATE char * tmp_put_file ARGS1(CONST char *, base)
{
    return tmpnam(NULL);
}


PRIVATE BOOL backup_mv ARGS2(CONST char *,	src,
			     CONST char *,	dest)
{
    static char * bname = NULL;

    FREE(bname);	/* From previous call */

    if (!src || !dest || !(bname=backup_name(dest)))
	return NO;

    if (rename(dest, bname) == -1) {
	CTRACE(stderr, "backup_mv: %s \"%s\" \"%s\" %s\n%s \"%s\"\n",
	       "Failed to make backup of old version (", dest, bname,
	       ")-- PUT fails.",
	       "Old version left intact, new is left in tmp file", src);
	return NO;
    }
    CTRACE(stderr,"backup_mv: \"%s\" backed up to \"%s\"\n",dest,bname);

    if (rename(src, dest) == -1) {
	CTRACE(stderr, "backup_mv: %s (mv \"%s\" \"%s\").\n%s \"%s\"\n",
	       "Failed to make new file current", src, dest,
	       "Trying to restore old version from", bname);
	if (rename(bname, dest) == -1) {
	    CTRACE(stderr,
		   "backup_mv: %s \"%s\" \"%s\" %s \"%s\" %s \"%s\"\n",
		   "Failed to restore old version (mv", bname, src,
		   ").\nPage becomes unavailable.\nBackup version is in",
		   bname, "and new version in", src);
	}
	else {
	    CTRACE(stderr, "backup_mv: %s\n%s \"%s\"\n",
		   "Managed to restore backup copy of old document.",
		   "New one is left in temporary file", src);
	}
	return NO;
    }
    CTRACE(stderr, "backup_mv: temp file \"%s\" renamed to \"%s\"\n",src,dest);
    return YES;
}


PRIVATE BOOL HTCommit ARGS2(HTRequest *,	req,
			    CONST char *, 	tmp_file)
{
    return backup_mv(tmp_file, HTReqTranslated);
}


PUBLIC int HTPut ARGS1(HTRequest *, req)
{
    char * tmp_file;
    FILE * fp;

    if (!req) return HT_LOADED;		/* What should I return?? */

    tmp_file = tmp_put_file(HTReqTranslated);
    if (!tmp_file) {
	CTRACE(stderr, "Unable to make temp PUT filename\n");
	return HT_LOADED;
    }

    fp = fopen(tmp_file, "w");
    if (!fp) {
	CTRACE(stderr,"Can't open tmp PUT file \"%s\" for writing\n",tmp_file);
	return HT_LOADED;
    }

    CTRACE(stderr, "Writing temporary PUT file \"%s\"\n", tmp_file);
    HTCopyBodyToFile(req, fp);
    fclose(fp);

    HTCommit(req, tmp_file);

    HTLoadError(req, 200, "Ok, file written to server");

    return HT_LOADED;
}


